# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Implementation of RAI Text Insights API."""

from responsibleai_text.rai_text_insights.rai_text_insights import \
    RAITextInsights

__all__ = ['RAITextInsights']
